#! /usr/local/bin/ruby -Ke

require "sdlcurses"
require "constant"
require "keyconvert"
require "utility"

class Client

  include Constant::Client
  include Constant::Object
  include Constant::Monster
  include Keyconvert
  include SDLcurses
  include Utility

  attr_accessor :player
  attr_reader :name

  def initialize( name )
    
    @name   = name
    @player = nil
   
    @screen = nil
       
    # H()Ȥäޥɤ
    # Server¦ǤϥݡȤʤΤǡClient¦ǥߥ졼Ȥޤ
    # ӤǤ

    # SDLcurses
    
    SDLcurses.initscr
    SDLcurses.noecho
    SDLcurses.cbreak
    
    # ѥͥꡣ

    @ranking = makepanel( DROWS, DCOLS, 0, 0 )
    @ranking.hide
    
    @mail    = makepanel( DROWS, DCOLS, 0, 0 )
    @mail.hide
    
    @ending  = makepanel( DROWS, DCOLS, 0, 0 )
    @ending.hide
    
    @message = makepanel( DROWS, DCOLS, 0, 0 )

    @left = makepanel( 1, DCOLS, 0, 0 )
    
    @main    = makepanel( MAXY + 3, DCOLS, 1, 0 )
    SDLcurses.wattroff( @main.window, A_BOLD )

    @store_message = nil
    @store_main    = nil
    
    @help = nil

    myupdate

  end

  def start_screen

    @start = makepanel( DROWS, DCOLS, 0, 0 )
    window = @start.window
    
    SDLcurses.wattron( window, A_BOLD | A_BLINK )
    SDLcurses.mvwaddstr( window, 8, 20, "RubyLarn 1.0.0" )
    SDLcurses.wattroff( window, A_BOLD | A_BLINK )
    SDLcurses.mvwaddstr( window, 10, 24,
			"This software is based on Larn 12.0" )
    SDLcurses.mvwaddstr( window, 11, 24,
			"Larn is copyrighted 1986 by Noah Morgan." )

    SDLcurses.wattron( window, A_BOLD )

    SDLcurses.mvwaddstr( window, 23, 0, "Enter: Start Game  Esc: Quit" )
    
    SDLcurses.wattroff( window, A_BOLD )
    
    myupdate

    exit if mygetc == ?\e
    
    @start.del

  end

  def configuration( loadable = nil )

    @start = makepanel( 25, DCOLS, 0, 0 )
    window = @start.window

    SDLcurses.wattron( window, A_BOLD )
    SDLcurses.mvwaddstr( window, 1, 1, "New Game" )

    10.times do | i |
      SDLcurses.wattron( window, A_REVERSE )
      SDLcurses.mvwaddstr( window, 3 + i, 3, i.to_s )
      SDLcurses.wattroff( window, A_REVERSE )
      SDLcurses.mvwaddstr( window, 3 + i, 4,
			  sprintf( ") Difficulty %d", i ) )
    end

    SDLcurses.mvwaddstr( window, 14, 1, "Load Game" )

    if loadable  
      SDLcurses.wattron( window, A_REVERSE )
      SDLcurses.mvwaddstr( window, 16, 3, "l" )
      SDLcurses.wattroff( window, A_REVERSE )
      SDLcurses.mvwaddstr( window, 16, 4,
			  ") You can load the game." )
    end
    
    SDLcurses.mvwaddstr( window, 23, 0, "Press 0 - 9 or l" )
    
    myupdate

    loop do

      input = mygetc
      
      if input.between?( ?0, ?9 ) || ( loadable && ( input == ?l ) ) ||
	  ( input == ?\e )
	@start.del
	return input
      end
	
    end
    
  end
  
  def myupdate
    Panel.update
    SDLcurses.doupdate
  end
  
  def makepanel( rows, cols, r_offset, c_offset )

 #   win = WINDOW.new( rows, cols, 0, 0 )
    
    win = WINDOW.new( rows, cols, r_offset, c_offset )
    SDLcurses.wattron( win, A_BOLD )
    
    panel = Panel::PANEL.new( win )
#    panel.move( r_offset, c_offset )

    SDLcurses.scrollok( win, true )
    # SDLcurses.keypad( win, true )
    
    panel

  end
  
  GO_DIRECT = {
    :magic_missile? => "+",
    :sonic_spear?   => "@",
    :fireball?      => "*",
    :cold?          => "O",
    :lightning?     => "~",
  }

  OMNI_DIRECT = {
    :cloud_kill?     => "'",
    :flood?          => "~",
    :magic_fire?     => "*",
#    :varporize_rock? => ".",
  }
  
  def magic_attack( spell, point )

    row, col = point
    s_time = 0

    window = @main.window
    SDLcurses.wattron( window, A_BLINK | A_BOLD )
    
    if ( k_v = GO_DIRECT.find{ | symbol, | spell.send( symbol ) } )
      
      SDLcurses.mvwaddch( window, row, col, k_v[ 1 ][ 0 ] ) 
      s_time = 0.08
      
    elsif ( k_v = OMNI_DIRECT.find{ | symbol, | spell.send( symbol ) } )

      # point濴ȤȾ1ϰˡˡ̤ȯ롣

      ( row - 1 ).upto( row + 1 ) do | r |
	( col - 1 ).upto( col + 1 ) do | c |
	  next if out_of_dungeon?( [ r, c ] )
	  SDLcurses.mvwaddch( window, r, c, k_v[ 1 ][ 0 ] )
	end
      end
      
      s_time = 4
      
    end

    SDLcurses.wattroff( window, A_BLINK | A_BOLD )
    
    cursor_to_player
    myupdate
    sleep( s_time )
    
  end
  
  def mygetc
    SDLcurses.flushinp
    SDLcurses.getch
  end

  def hide_main

    @left.hide
    @main.hide
    SDLcurses.redrawwin( @message.window ) if $REDRAW
    myupdate
    mygetc
    @left.show
    @main.show
    myupdate

  end

  def multi_move_check

    # ¿ϤꤵƤʤޤϲλ鵶
    return false if !@multi_com || ( ( @multi_times -= 1 ) < 0 )

    # ᤯˥󥹥뤫ɤ?
    return true unless @screen

    p_point = @screen.players.keys[ 0 ]
    
    if @screen.monsters.find { | m_point, | distance( p_point, m_point ) <= 4 }
      @multi_times = 0
      return false
    end

    return true unless DIRS.include?( @multi_com )
    
    # Ԥ˥֥ȤˤϡȤꤢߤޤ餻롣
    n_point = next_point( p_point, @multi_com )
    
    if @screen.objects.key?( n_point )
      @multi_times = 0
      return false
    end

    # ϷŪ˻ߤޤäɤ硣ƻʬʤɡ

    if better_to_stop?( p_point, n_point, @multi_com )
      @multi_times = 0
      return false
    end
    
    true
    
  end

  # ʲΤ褦ʾˤϡ¿ưߤ롣
  # ǤʤˤʤळȤΤǡ
  
  # 
  # 
  #   

  # Τ褦ʾˤ³ԡ
  # ̵롣
  
  #
  # 
  #     
  #   

  # ξˤ³ԡʤɤʤϤǤ

  #
  #    
  #   
  # 

  DIR2SIDE = {
    ?h => [ ?y, ?b ],
    ?j => [ ?b, ?n ],
    ?k => [ ?y, ?u ],
    ?l => [ ?u, ?n ],
    ?y => [ ?h, ?k ],
    ?u => [ ?k, ?l ],
    ?b => [ ?h, ?j ],
    ?n => [ ?l, ?j ],
  }

  NARUTO2SIDE = {
    ?h => [ ?k, ?j ],
    ?j => [ ?h, ?l ],
    ?k => [ ?h, ?l ],
    ?l => [ ?k, ?j ],
    ?y => [ ?u, ?b ],
    ?u => [ ?y, ?n ],
    ?b => [ ?y, ?n ],
    ?n => [ ?u, ?b ],
  }

  def can_move_to_dirs?( point, dirs )
    dirs.find { | dir | can_move_to?( next_point( point, dir ) ) } 
  end

  
  def better_to_stop?( from, to, dir )
    
    return true  unless can_move_to?( to )
    return false if can_move_to_dirs?( from, NARUTO2SIDE[ dir ] )

    can_move_to_dirs?( from, DIR2SIDE[ dir ] )

  end

  def can_move_to?( point )
    !out_of_dungeon?( point ) && ( @screen.maze[ point[ 0 ] ][ point[ 1 ] ] != OWALL )
  end
  
  def distance( from, to )
    [ ( from[ 0 ] - to[ 0 ] ).abs, ( from[ 1 ] - to[ 1 ] ).abs ].max
  end

  def time_left( left )

    window = @left.window
    
    SDLcurses.wmove( window, 0, 0 )

    left = 99999 if left > 99999
    SDLcurses.mvwaddstr( window, 0, 0,
			sprintf( "%-45d x: Look     5: Rest    ?: Help", left ) )
    
    cursor_to_player

    myupdate

    
  end
  
  # ޥɤϤ¥
  def command

    cursor_to_player
    myupdate
    
    return @multi_com if multi_move_check

    loop do

      com = mygetc
      
      if com == ?\t
	hide_main
	next
      end

      if MULTI_MOVE.include?( com )
	@multi_com   = multi2one( com )
	@multi_times = 100
	return @multi_com
      elsif TEN_KEY.include?( com )
	return( tenkey2hjklyubn( com ) )
      else
	return com
      end
      
    end
            
  end

  def test 

  end
  
  def newcavelevel( screen, volcano )
    @screen = screen
    @volcano = volcano
    @multi_times = 0
  end
  
  def drawscreen( screen, to_update = true ) # 󥸥ʬ衣

    SDLcurses.werase( @main.window )

    0.upto( MAXY - 1 ) do | row |
      0.upto( MAXX - 1 ) do | col |
	showpoint( screen, row, col )
      end
    end

    cursor_to_player
    myupdate if to_update
    
  end

  def cursor_to_player

    return unless @screen
    return unless ( point = @screen.players.keys[ 0 ] )
    
    SDLcurses.wmove( @main.window, *point )
        
  end

  
  # ( row, col )濴ˤơȾradiusΥꥢԤ
  def showcell( screen, point, radius, update_flag = true )
    
    row, col = point

    # drawscreen( @screen, false ) if @screen

    # Ncurses.wattron( @main.window, A_BOLD )

    ( row - radius ).upto( row + radius ) do | r |
      next if ( r < 0 ) || ( r > MAXY - 1 )
      ( col - radius ).upto( col + radius ) do | c |
	next if ( c < 0 ) || ( c > MAXX - 1 )
	showpoint( screen, r, c )
      end
    end

    # Ncurses.wattroff( @main.window, A_BOLD )

    cursor_to_player
      
  end

  def showpoint( screen, row, col )

    # ͥ٤ϡ
    # 1: 󥹥ץ쥤䡼
    # 2: ֥
    # 3: ϩ
    # ν硣

    point  = [ row, col ]
    window = @main.window
    
    if ( maze = screen.maze[ row ][ col ] ) == OWALL
      SDLcurses.mvwaddch( window, row, col, OBJCHAR[ maze ] )
    elsif mon = screen.monsters[ point ]
      SDLcurses.wattron( window, A_BOLD )
      SDLcurses.mvwaddch( window, row, col, MONCHAR[ mon ] )
      SDLcurses.wattroff( window, A_BOLD )
    elsif player = screen.players[ point ]
      SDLcurses.mvwaddch( window, row, col, ?@ )
    elsif obj = screen.objects[ point ]
      SDLcurses.wattron( window, A_BOLD | A_BLINK )
      SDLcurses.mvwaddch( window, row, col, OBJCHAR[ obj ] )
      SDLcurses.wattroff( window, A_BOLD | A_BLINK )
    elsif sphere = screen.spheres[ point ]
      SDLcurses.wattron( window, A_REVERSE )
      SDLcurses.mvwaddch( window, row, col, sphere )
      SDLcurses.wattroff( window, A_REVERSE )
    else
      SDLcurses.mvwaddch( window, row, col, ?\s )
    end

  end

  def close
    SDLcurses.endwin
  end

  def status_routine( row, str )
    SDLcurses.wattron( @main.window, A_BLINK | A_BOLD | A_UNDERLINE )
    SDLcurses.mvwaddstr( @main.window, row, 0, sprintf( "%-80s", str ) )
    SDLcurses.wredrawln( @main.window, row, 1 ) if $REDRAW
    SDLcurses.wattroff( @main.window, A_BLINK | A_BOLD | A_UNDERLINE )
  end
  
  def status1( format1, status_array )
    status_routine( MAXY, sprintf( format1, *status_array ) )
  end

  def status2( format2, status_array )
    status_routine( MAXY + 1, sprintf( format2, *status_array ) )
  end

  def my_scroll( panel, num = 1 )

    SDLcurses.wscrl( panel.window, num )
    SDLcurses.redrawwin( panel.window ) if $REDRAW
  
  end
  
  def m_scroll( num = 1 )
    
    SDLcurses.wscrl( @message.window, num )
    SDLcurses.wredrawln( @message.window, MAXY + 2, DROWS - MAXY - 2 ) if $REDRAW

  end

  def version( arg )

    m_scroll( arg.size )
    window = @message.window
    
    arg.each_with_index do | line, i |
      SDLcurses.mvwaddstr( window, DROWS - arg.size + i, 0, line )
    end
    
    SDLcurses.redrawwin( window ) if $REDRAW
    
    cursor_to_player
    myupdate
        
  end

  IFORMAT1 = "%s)   %s"
  IFORMAT2 = " + %-2d"
  IFORMAT3 = " - %-2d"
  IFORMAT4 = " (%s)"
  MFORMAT  = "%s)   %d %s"

  def format2str( iformat )

    if iformat[ 0 ] == "."
      return( sprintf( MFORMAT, iformat[ 0 ], iformat[ 2 ], iformat[ 1 ] ) )
    end
    
    str = sprintf( IFORMAT1, *iformat )

    if iformat[ 2 ] && ( iformat[ 2 ] != 0 )
      str << sprintf( ( iformat[ 2 ] < 0 ) ? IFORMAT3 : IFORMAT2,
		     iformat[ 2 ].abs )
    end

    if iformat[ 3 ]
      str << sprintf( IFORMAT4, iformat[ 3 ] )
    end

    str
      
  end
  
  def item( item_format )
    message( format2str( item_format ) )
  end

  def select( select_mesg, list, press_space, abort )

    message( select_mesg )

    loop do
      
      if ( ( com = mygetc ) == ?* ) || ( com == ?D ) 
	if ( char = itemlist( list, press_space ) ) ; return char ; end
	next
      end

      if com == ?\e
	SDLcurses.waddstr( @message.window,  abort )
	cursor_to_player
	myupdate
      end
	
      return com
      
    end

  end

  def write_press_space( window, row, press_space )

    SDLcurses.wmove( window, row, 0 )
    
    press_space.each_with_index do | str, i |
      SDLcurses.wattron( window, A_REVERSE ) if i == 1
      SDLcurses.waddstr( window, str ) if str 
      SDLcurses.wattroff( window, A_REVERSE ) if i == 1
    end
    
  end

  def itempage( panel, to_write, press_space )

    # WINDOWޤ뤴ȼؤܸβʤ?
    panel.replace( win = WINDOW.new( to_write.size + 1, DCOLS, 0, 0 ) )
    SDLcurses.wattron( win, SDLcurses::A_BOLD )
    SDLcurses.redrawwin( win ) if $REDRAW
    
    to_write.each_with_index do | string, i |
      SDLcurses.mvwaddstr( win, i, 0, string )
    end
    
    write_press_space( panel.window, to_write.size, press_space )
    myupdate

    ( com = mygetc ).between?( ?a, ?z ) ? com : nil

  end
    
  def itemlist( iarray, press_space, time = nil )

    # item_arrayΥϡ⡹ ( 26 + 1 )

    to_write = iarray.collect{ | format | format2str( format ) }
    to_write.push( time ) if time

    first  = ( ( to_write.size > DROWS - 1 )? DROWS - 1 : to_write.size )
    second = to_write.size - first 

    panel = makepanel( DROWS, DCOLS, 0, 0 )
    
    if com = itempage( panel, to_write[ 0, first ], press_space )
      panel.del
      myupdate
      return( com ) 
    end
    
    if ( second > 0 )
      com = itempage( panel, to_write[ first, second ], press_space )  
    end

    panel.del
    myupdate

    com
    
  end

  def all_item( item_array, time_format, elapsed, left, press_space )

    itemlist( item_array, press_space, sprintf( time_format, elapsed, left ) )

  end
    
  def message( str, cursor = true )

    m_scroll
    SDLcurses.mvwaddstr( @message.window,  DROWS - 1, 0, str ) if str

    multi_cancel
    SDLcurses.redrawwin( @message.window ) if $REDRAW
    
    cursor_to_player if cursor
    myupdate
    
  end

  def investigate_message( str )
    message( str, false )
  end

  
  def statue_message( str )

    m_scroll( 2 )
    multi_cancel
    SDLcurses.mvwaddstr( @message.window, DROWS - 1, 0, str ) if str
    cursor_to_player
    myupdate
        
  end

  alias_method( :mirror_message, :statue_message )
  alias_method( :pit_message,    :statue_message )
  
  def s_scroll
    my_scroll( @store_message )
  end
  
  def store_message( str )

    s_scroll
    SDLcurses.mvwaddstr( @store_message.window, DROWS - 1, 0, str ) if str

    myupdate
    
  end

  def add_message( str )

    SDLcurses.waddstr( @message.window, str ) if str
    cursor_to_player
    myupdate

  end

  def reach_question( find_str, select_str, choice )
    
    m_scroll( 3 )
    window = @message.window
    
    SDLcurses.mvwaddstr( window, DROWS - 2, 0, find_str )
    SDLcurses.mvwaddstr( window, DROWS - 1, 0, select_str )
    myupdate
    
    multi_cancel
    
    loop do
      
      next unless choice.key?( com = mygetc )
      
      SDLcurses.waddstr( window, choice[ com ] )
      myupdate

      return com

    end
    
  end
  
  def question( str, choice )

    message( str )

    loop do
      
      next unless choice.key?( com = mygetc )

      SDLcurses.waddstr( @message.window, choice[ com ] )
      myupdate

      return com

    end
    
  end

  def quit_question( question, answer, second_question )

    message( question )

    loop do

      if answer.key?( com = mygetc )
	SDLcurses.waddstr( @message.window,  answer[ com ] )
	myupdate

	return com

      end

      m_scroll

      win = @message.window
      win.move( DROWS - 1, 0 )

      second_question.each_with_index do | string, i |
	SDLcurses.wattron( win, A_REVERSE ) if i == 0 || i == 2 
	SDLcurses.waddstr( win, string )
	SDLcurses.wattroff( win, A_REVERSE ) if i == 0 || i == 2
      end

      myupdate
      
    end
    
  end

  def home( home_mesg, press_mesg, modulo )

    panel  = makepanel( DROWS, DCOLS, 0, 0 )
    window = panel.window

    SDLcurses.wattron( window, A_BOLD )
    
    home_mesg.each_with_index do | mesg, i |
      next if !mesg || mesg.size == 0
      SDLcurses.mvwaddstr( window, i, 0,
		  sprintf( mesg, ( mesg =~ /%d/ ) ? modulo : @name ) )
      
    end

    SDLcurses.wmove( window, home_mesg.size, 0 )
    
    press_mesg.each_with_index do | mesg, i |
      next if !mesg || mesg.size == 0
      SDLcurses.wattron( window, A_REVERSE ) if ( i == 1 ) || ( i == 3 )
      SDLcurses.waddstr( window, mesg )
      SDLcurses.wattroff( window, A_REVERSE ) if ( i == 1 ) || ( i == 3 )
    end

    myupdate

    loop { break if mygetc == ESC }

    panel.del
    myupdate
    
  end

  def gold( mesg, amount )

    m_scroll( 3 )
    
    mesg.each_with_index do | str, i |
      SDLcurses.mvwaddstr( @message.window, DROWS - 2 + i, 0,
			  sprintf( str, amount ) )
    end
    
    multi_cancel
    myupdate
      
  end

  def kick_damage( str, ndamage )

    message( sprintf( str, ndamage ) )
    myupdate

  end

  def learn_spell( format, code, name )
    message( sprintf( format, code, name ) )
  end

  def multi_cancel
    @multi_com  = nil
    @multi_times = 0
  end

  def opening( mesg_array, press )
    
    panel  = makepanel( DROWS, DCOLS, 0, 0 )
    window = panel.window

    SDLcurses.wattron( window, SDLcurses::A_BOLD )
    
    mesg_array.each_with_index do | str, i |
      SDLcurses.mvwaddstr( window, i, 0, str )
    end

    SDLcurses.wmove( window, DROWS - 1, 0 )

    press.each_with_index do | str, i |
      SDLcurses.wattron( window, SDLcurses::A_REVERSE ) if i == 1
      SDLcurses.waddstr( window, str )
      SDLcurses.wattroff( window, SDLcurses::A_REVERSE ) if i == 1
    end

    
    myupdate
    mygetc

    panel.del
    myupdate

  end

  BACKSPACE =   8
  ESC       =  27
  DELETE    = 127

  def altar_get_amount( mesg ) ; get_amount( @message, mesg ) ; end

  alias_method( :drop_get_amount,  :altar_get_amount )
  alias_method( :debug_get_amount, :altar_get_amount )

  
  # ۤϤ롣
  def get_amount( panel, mesg )

    if panel == @message
      message( mesg )
    else
      store_message( mesg )
    end

    window = panel.window
    
    # ncurses-ruby饤֥ϡgetyxΰμ߷פѡ
    window.getyx( row = [], col = [] )

    amount = 0
    
    loop do

      com = mygetc
    
      if com == ESC # ϥ󥻥롣
	return( 0 )
      elsif com == BACKSPACE || com == DELETE
	amount = ( amount / 10 )
	SDLcurses.wmove( window, row[ 0 ], col[ 0 ] )
	
	SDLcurses.wclrtoeol( window )
	SDLcurses.waddstr( window, amount.to_s )  if amount > 0
	myupdate
      elsif com.between?( ?0, ?9 )
	next if amount > 1000000
	amount = amount * 10 + com - ?0
	SDLcurses.waddstr( window,  com.chr )
	myupdate
      elsif com == ?\n || com == ?\r
	return amount
      end
      
    end

  end

  def get_direction( mesg )

    message( mesg )
    
    # ѤϤޤǡ֤
    loop do
       if DIRS.include?( dir = mygetc )
	 return( dir )
       elsif TEN_KEY.include?( dir )
	 return tenkey2hjklyubn( dir )
       end
    end

  end
  
  def get_spell( enter_spell, aborted, header, list, press_space )

    message( enter_spell )

    spell = ""
    
    3.times do | i |

      input = mygetc

      if ( i == 0 ) && ( input == ?D )
	show_found_list( header, list, press_space )
	redo
      elsif input == ?\e
	add_message( aborted )
	return nil
      elsif input.between?( ?a, ?z ) || input.between?( ?A, ?Z )
	spell << input.chr
      end
      
    end
    
    spell
    
  end

  # 3ȤߤˤȤɬפʥ

  # [ "", "ǥ", "å", "", "륿" ]ȡ

  #     ǥ   å
  #     륿

  # Ȥˡ5 / 3 + 1 = 2 ɬפˤʤޤ
  
  def three_column_rows( size )
    size / 3 + ( ( ( size % 3 ) == 0 ) ? 0 : 1 )
  end

  def entry_to_string( entry )

    if entry.is_a?( String )
      entry
    elsif entry.is_a?( Array ) && ( entry.size == 2 )
      sprintf( "%s %s", *entry )
    end

  end

  SPELL_OFFSET = [ 0, 27, 54 ]
    
  def show_found_list( header, list, press_space, last_update = true )

    # list:
    # list = [ "enchant armor", "enchant weapon", ....,  ]

    rows = 4 + three_column_rows( list.size )

    panel = makepanel( rows, DCOLS, 0, 0 )
    win   = panel.window

    SDLcurses.wattron( win, A_BOLD | A_BLINK )
    SDLcurses.redrawwin( win ) if $REDRAW
    
    # Ǿʤ˥å
    SDLcurses.mvwaddstr( win, 0, 0, header )

    # ꥹȤȤߤɽ
    list.each_with_index do | entry, i |
      SDLcurses.mvwaddstr( win, 2 + i/3, SPELL_OFFSET[ i % 3 ],
			entry_to_string( entry ) )
    end

    SDLcurses.wattroff( win, A_BLINK )
    
    # ǲʤ press space to continue
    write_press_space( win, rows - 1, press_space )

    myupdate

    mygetc

    panel.delete

    myupdate if last_update
    
  end

  # Ի󥹥Ρե٥åȥɤϤ롣
  def get_genocide( mesg )

    message( mesg )

    loop do
      
      char = mygetc
      next if !( char.between?( ?a, ?z ) ) && !( char.between?( ?A, ?Z ) )

      SDLcurses.waddstr( @message.window, char.chr )
      myupdate

      return char

    end

  end

  # ߤޤǤȯ̤ˡʪΥꥹȤ
  # ֤ɽޤ
  
  def all_item_found( m_header, m_list, s_header, s_list, p_header, p_list,
		     press_space )

    # header:
    # ʸ"The magic spells you have discovered thus far:"ʤɡ

    # list: ˡξ硣
    # ηΡޤǤ
    # [ [ "pro",         "protection"    ],
    #   [ "mle",         "magic missile" ],
    #   [ ˡΥɡ ˡ      ], ] 

    # list: ʪξ
    # [ "enchant armor", "echantweapon", ... ]
    
    show_found_list( m_header, m_list, press_space, false ) # ˡ
    show_found_list( s_header, s_list, press_space, false ) # ʪ
    show_found_list( p_header, p_list, press_space, false ) # 

  end

  def tradepost( mesg )
    makestore( 7, mesg )
  end

  def tradepost_question( mesg )

    s_scroll

    win = @store_message.window

    SDLcurses.wmove( win, DROWS - 1, 0 )
    
    mesg.each_with_index do | string, i |
      SDLcurses.wattron( win, A_REVERSE ) if i == 2 || i == 4
      SDLcurses.waddstr( win, string )
      SDLcurses.wattroff( win, A_REVERSE ) if i == 2 || i == 4
    end

    myupdate

    loop do
      com = mygetc
      return ?* if com == ?D
      return com if com == ?\e || com == ?* || com.between?( ?a, ?z )
    end

  end
  
  def store_item( item_format )
    store_message( format2str( item_format ) )
  end

  alias_method( :tradepost_item, :store_item )
  alias_method( :dndstore_item,  :store_item )

  def tradepost_sell( mesg, answer_hash )

    tradepost_message( mesg )

    window = @store_message.window
    
    loop do 
      next unless answer_hash.key?( com = mygetc )

      SDLcurses.waddstr( window, answer_hash[ com ] )
      return com

    end

  end

  def store_exit
    
    SDLcurses.curs_set( 1 )

    @store_main.delete
    @store_message.delete

    myupdate
    
  end

  alias_method( :tradepost_exit, :store_exit )
  alias_method( :dndstore_exit,  :store_exit )
  alias_method( :school_exit,    :store_exit )
  
  alias_method( :tradepost_message, :store_message )
  alias_method( :dndstore_message,  :store_message )
  alias_method( :school_message,    :store_message )
  
  def makestore( main_rows, mesg )

    SDLcurses.curs_set( 0 )
    
    @store_message = makepanel( DROWS,     DCOLS, 0, 0 ) 
    @store_main    = makepanel( main_rows, DCOLS, 0, 0 )
    
    window = @store_main.window

    SDLcurses.wattron( window, A_BOLD )
    SDLcurses.wattron( @store_message.window, A_BOLD )
    
    mesg.each_with_index do | string, i |
      SDLcurses.mvwaddstr( window, i, 0, string ) if string
    end
    
  end

  def dndstore_tax( tax_mesg, press_escape )

    window = @store_main.window
    
    tax_mesg.each_with_index do | string, i |
      SDLcurses.mvwaddstr( window, 4 + i, 0, string )
    end

    myupdate

    store_standout_write( press_escape, [ 1 ] )
    mygetc
    
  end

  def dndstore( mesg )
    makestore( 18, mesg )
  end

  def dndstore_money( gold )
    SDLcurses.mvwaddstr( @store_main.window, 17, 50, gold )
  end

  def dndformat2str( dndformat )
    sprintf( "%c) %-28s%6s", *dndformat )
  end
  
  def dndstore_choose( list, mesg )

    window = @store_main.window
    
    # ľΥꥹȤõ
    13.times do | i |
      SDLcurses.wmove( window, i + 4, 0 )
      SDLcurses.wclrtoeol( window )
    end

    list.each_with_index do | dndformat, i |
      next unless dndformat
      SDLcurses.mvwaddstr( window, i / 2 + 4, ( i % 2 ) * 40,
			dndformat2str( dndformat ) )
    end

    SDLcurses.redrawwin( window ) if $REDRAW
        
    store_standout_write( mesg, [ 2, 4 ] )

    loop do
      com = mygetc
      return( com ) if com == ?\e || com == ?\s || com.between?( ?a, ?z )
    end
        
  end

  alias_method( :tradepost_all_item, :all_item )

  def school( welcome, cost )
    makestore( 18, welcome )
    SDLcurses.mvwaddstr( @store_main.window, 15, 16, cost )
  end

  def store_standout_write( mesg, to_standout = [] )

    s_scroll

    win = @store_message.window
    SDLcurses.wmove( win, DROWS - 1, 0 )

    mesg.each_with_index do | string, i |
      SDLcurses.wattron( win, A_REVERSE ) if to_standout.include?( i )
      SDLcurses.waddstr( win, string )
      SDLcurses.wattroff( win, A_REVERSE ) if to_standout.include?( i )
    end

    myupdate
    
  end  
  
  def school_course( header, course, mesg )

    s_win = @store_main.window
    
    SDLcurses.mvwaddstr( s_win, 4, 20, header )
       
    course.each_with_index do | string, i |
      SDLcurses.wmove( s_win, 6 + i, 16 )
      SDLcurses.wclrtoeol( s_win )
      SDLcurses.waddstr( s_win, string ) if string
    end
    
    # 939 "What is your choice "
    # 940 "["
    # 941 "escape" # ȿž
    # 942 " to leave"
    # 943 "] ? "

    store_standout_write( mesg, [ 2 ] )

    loop do
      com = mygetc
      return( com ) if com == ?\e || com.between?( ?a, ?h )
    end
    
  end

  def school_money( money )
    SDLcurses.mvwaddstr( @store_main.window, 17, 30, money )
    myupdate
  end

 
  # lrs  
 
   def lrs( mesg )
    makestore( 6, mesg )
  end

  def lrs_tax( tax_string )

    window = @store_main.window
    
    SDLcurses.wmove( window, 3, 0 )
    SDLcurses.wclrtoeol( window )
    SDLcurses.mvwaddstr( @store_main.window, 3, 0, tax_string )
    SDLcurses.wredrawln( window, 3, 0 ) if $REDRAW
  
    myupdate
    
  end

  def lrs_money( money_string )

    window = @store_main.window
    
    SDLcurses.wmove( window, 5, 0 )
    SDLcurses.wclrtoeol( window )
    SDLcurses.mvwaddstr( @store_main.window, 5, 0, money_string )
    SDLcurses.wredrawln( window, 5, 0 ) if $REDRAW
    
    myupdate
    
  end

  def lrs_command( mesg, command_hash )

    # mesg:    
    # 972 "Your wish? "
    # 973 "["
    # 974 ("
    # 975 "p" # Ĵɽ
    # 976 ")"
    # 977 " pay taxes, or "
    # 978 "escape" # Ĵɽ
    # 979 "]  "

    store_standout_write( mesg, [ 3, 6 ] )
    window = @store_message.window
    
    loop do

      if ( com = mygetc ) == ?\t
	store_main_hide
	next
      end
      
      next unless command_hash.key?( com )

      SDLcurses.waddstr( window, command_hash[ com ] )
      myupdate

      return( com )

    end

  end

  alias_method( :lrs_message, :store_message )
  alias_method( :lrs_exit, :store_exit )

    
  def lrs_pay_taxes( how_much )
    get_amount( @store_message, how_much )
  end

  def store_main_hide

    @store_main.hide
    myupdate

    mygetc
    
    @store_main.show
    myupdate
    
  end

  # bank
  
  def bank( mesg )
    makestore( 18, mesg )
  end

  alias_method( :bank_exit,    :store_exit )
  alias_method( :bank_message, :store_message )

  def bank_pay_taxes( mesg, press_escape )

    # 1181 "Press "
    # 1182 "escape"
    # 1183 " to leave: "

    window = @store_main.window
    
    mesg.each_with_index do | string, i |
      SDLcurses.mvwaddstr( window, 2 + i, 0, string )
    end

    myupdate

    store_standout_write( press_escape, [ 1 ] )

    mygetc

  end

  def bank_bankaccount( bankaccount )
    SDLcurses.mvwaddstr( @store_main.window, 16, 30, bankaccount )
    myupdate
  end

  def bank_gold( gold )
    SDLcurses.mvwaddstr( @store_main.window, 17, 39, gold )
    myupdate
  end

  def gemformat2str( gemformat )
    sprintf( "%c) %-28s %7d ", *gemformat )
  end
  
  
  def bank_gemlist( header, gemlist )

    win = @store_main.window
    
    SDLcurses.mvwaddstr( win, 2, 0, header )

    gemlist.each_with_index do | gemformat, i |
      SDLcurses.wmove( win, 3 + i/2, 40 * ( i % 2 ) )
      SDLcurses.wclrtoeol( win )
      SDLcurses.waddstr( win, gemformat2str( gemformat ) ) if gemformat
    end

    myupdate
        
  end

  def bank_command( command_hash, you_wish )

    # you_wish:
    
    # 1190 "Your wish? "
    # 1191 "["
    # 1192 "("
    # 1193 "d" # ȿž
    # 1194 ")"
    # 1195 " deposit, "
    # 1196 "("
    # 1197 "w" # ȿž
    # 1198 ")"
    # 1199 " withdraw, "
    # 1200 "("
    # 1201 "s" # ȿž
    # 1202 ")"
    # 1203 " sell a stone, or "
    # 1204 "escape" # ȿž
    # 1205 "]  "

    store_standout_write( you_wish, [ 3, 7, 11, 14 ] )
    window = @store_message.window
    
    loop do

      next unless command_hash.key?( com = mygetc )
      SDLcurses.waddstr( window, command_hash[ com ] )
      myupdate

      return( com )

    end
    
  end

  def bank_get_amount( mesg )
    get_amount( @store_message, mesg )
  end

  def bank_which_stone( which_stone )
    store_message( which_stone )

    loop do
      com = mygetc
      return( com ) if com == ?\e || com.between?( ?a, ?z )
    end

  end

  def side_status( status_array )

    window = @main.window

    SDLcurses.wattron( window, A_BOLD )
    
    status_array.each_with_index do | string, i |
      SDLcurses.wmove( window, i, 68 )
      SDLcurses.wclrtoeol( window )
      SDLcurses.waddstr( window, string ) if string 
    end

    SDLcurses.wattroff( window, A_BOLD )
    
    cursor_to_player
    myupdate

  end

  def help( first, second, third, fourth, fifth, press_space, press_enter )

    @help = makepanel( DROWS, DCOLS, 0, 0 )
    help_write( first,  press_space, [ 1, 3 ] )
    help_write( second, press_space, [ 1, 3 ] )
    help_write( third,  press_space, [ 1, 3 ] )
    help_write( fourth, press_space, [ 1, 3 ] )
    help_write( fifth,  press_enter, [ 1 ]    )

    @help.del
    myupdate
    
  end

  def help_write( mesg_array, press_space, to_standout = [] )

    win = @help.window
    SDLcurses.werase( win )
    
    SDLcurses.wattron( win, SDLcurses::A_BOLD )
    
    mesg_array.each_with_index do | string, i |
      SDLcurses.wstandout( win ) if i == 0
      SDLcurses.mvwaddstr( win, i,
			  ( i == 0 )? ( DCOLS - string.size ) / 2 :  0,
			string )
      SDLcurses.wstandend( win ) if i == 0
    end

    SDLcurses.wmove( win, DROWS - 1, 0 )

    press_space.each_with_index do | string, i |
      SDLcurses.wstandout( win ) if to_standout.include?( i )
      SDLcurses.waddstr( win, string )
      SDLcurses.wstandend( win ) if to_standout.include?( i )
    end

    SDLcurses.redrawwin( win ) if $REDRAW
    myupdate

    mygetc

  end

  def happy_ending( mesg )

    @ending.show
    myupdate

    window = @ending.window
    
    mesg.each_with_index do | string, i |
      SDLcurses.mvwaddstr( window, i, 0, string )
      myupdate
      sleep( 8 ) if string.size > 0
    end

    mygetc
    
  end

  alias_method( :timeover_with_the_potion,    :happy_ending )
  alias_method( :timeover_without_the_potion, :happy_ending )

  def ranking_latest( latest_format, latest_arg )

    @ranking.show
    
    # ǿλ˴ǡ
    SDLcurses.wattron( @ranking.window, A_BOLD )
        
    SDLcurses.mvwaddstr( @ranking.window, 0, 0,
		      sprintf( latest_format, *latest_arg ) )

    
  end

  def ranking_winner( *arg )

    # ԥ٥8
    ranking_routine( 2, *arg )

  end

  def ranking_routine( offset, header, format, arg_array, empty_string,
		      to_stand_score = nil )

    win = @ranking.window
    
    SDLcurses.mvwaddstr( win, offset, 0, header )

    if arg_array.empty?
      SDLcurses.mvwaddstr( win, offset + 1, 0, empty_string )
    end
    
    arg_array.each_with_index do | arg, i |
      score  = arg[ 0 ].to_i
      string = sprintf( format, *arg )

      if score == to_stand_score
	string << ( " " * ( DCOLS - string.size ) )
	SDLcurses.wattron( win, A_REVERSE )
      end

      SDLcurses.mvwaddstr( win, offset + 1 + i, 0, string )

      if score == to_stand_score
	SDLcurses.wattroff( win, A_REVERSE )
	to_stand_score = nil
      end

    end

    myupdate
    
  end
  
  def ranking_loser( *arg )
    
    # Լԥ٥8
    ranking_routine( 12, *arg )
    SDLcurses.wmove( @ranking.window, DROWS - 1, 0 )
    myupdate
    mygetc
    mygetc
    
  end

  def killed
    sleep 4
  end

  def get_point

    # ruby-ncursesΡgetyxμ
    y = []
    x = []

    window = @main.window
    window.getyx( y, x )

    from = [ y[ 0 ], x[ 0 ] ]
    
    loop do

      if ( com = mygetc ) == ?\e
	return nil
      elsif DIRS.include?( com ) || TEN_KEY.include?( com )
	com = tenkey2hjklyubn( com ) if TEN_KEY.include?( com )
	
	next if out_of_dungeon?( to = next_point( from, com ) )

	SDLcurses.wmove( window, *to )
	myupdate
	return to

      end

    end
    
  end

  def out_of_dungeon?( point )
   !( point[ 0 ].between?( 0, MAXY - 1 ) &&
     point[ 1 ].between?( 0, MAXX - 1 ) )
  end

  def mail( mail_string_array )

    win = @mail.window

    SDLcurses.werase( win )
    SDLcurses.wattron( win, SDLcurses::A_BOLD )
    
    mail_string_array.each_with_index do | string, i |
      SDLcurses.mvwaddstr( win, i, 0, string )
    end

    SDLcurses.redrawwin( win ) if $REDRAW
    @mail.show
    myupdate
    mygetc

  end
  
end
