#! /usr/local/bin/ruby -Ke

require "cavelevel"
require "doublearray"
require "constant"
require "display"
require "utility"
require "wizard"
require "scroll"
require "potion"
require "store"
require "trap"
require "command"
require "regenerate"
require "movemonster"
require "spell"
require "death"
require "distancemap"

# @know: 볦
# @objects: ޤǤȯƥࡢüϷ
# @monsters ޤǤȯ󥹥
# 󥹥ưΤǡºݤΰ֤Ȥϰ㤦⤢롣

class Screen

  include Constant::Cavelevel
  
  attr_accessor :maze, :objects, :monsters, :players, :spheres
  
  def initialize
    @maze     = Doublearray.new( MAXY, MAXX )
    @objects  = {}
    @monsters = {}
    @players  = {}
    @spheres  = {}
  end

  def know?( row, col )
    @maze[ row ][ col ]
  end

  def know_object?( point, object )
    ( id = @objects[ point ] ) && ( id == object.id )
  end

  def showcell( player, cave, point, radius )
    showcell_area( player, cave, point, radius, radius )
  end

  def showcell_area( player, cave, point, r_radius, c_radius )

    row, col = point

    # ܤλˤϡץ쥤䡼ΰ֤Τ߹롣
    # Ƥ櫤ɽʤ
    # ܤ˸ʤ󥹥ϡƥõ뤳Ȥɽ롣

    ( row - r_radius ).upto( row + r_radius ) do | r |
      next if ( r < 0 ) || ( r > MAXY - 1 )
      ( col - c_radius ).upto( col + c_radius ) do | c |
	next if ( c < 0 ) || ( c > MAXX - 1 )

	point = [ r, c ]

	# ץ쥤䡼ι
	if another = cave.point2player( point )
	  @players[ point ] = another.id
	else
	  @players.delete( point )
	end

	next if player.blindcount?
		
	# ϩʬι
	@maze[ r ][ c ] = cave.maze[ r ][ c ]
	
	# ֥Ȥι
	if object = cave.point2object( point )
	  @objects[ point ] = object.id unless object.hidden? 
	else 
	  @objects.delete( point )
	end
	
	# 󥹥ι
	if monster = cave.point2monster( point )
	  if player.can_see_monster?( monster )
	    @monsters[ point ] = monster.id
	  else
	    @objects.delete( point )
	  end
	else
	  @monsters.delete( point )
	end

	if sphere = cave.point2sphere( point )
	  @spheres[ point ] = sphere.dir
	else
	  @spheres.delete( point )
	end
	
      end

    end

  end # def showcell

  def showall( player, cave )

    @maze   = cave.maze.dup

    @objects.clear
    cave.objects.each do | point, object |
      @objects[ point ] = object.id unless object.hidden?
    end

    monster_detection( player, cave )

    @players.clear
    cave.players.each { | point, player | @players[ point ] = player.id }
    
  end # def showall

  def gold_detection( player, cave )

    return if player.blindcount?

    cave.objects.each do | point, object |
      @objects[ point ] = object.id if object.for_gold_detection?
    end

  end

  def treasure_finding( player, cave )

    return if player.blindcount?

    cave.objects.each do | point, object |
      @objects[ point ] = object.id if object.for_treasure_finding?
    end

  end
  
  def monster_detection( player, cave )

    return if player.blindcount?

    @monsters.clear
    
    cave.monsters.each do | point, monster |

      if player.can_see_monster?( monster )
	@monsters[ point ] = monster.id
      else
	@objects.delete( point )
      end

    end
    
  end

  def forgetfulness( player, cave )

    @maze     = Doublearray.new( MAXY, MAXX )
    @objects.clear
    @monsters.clear
   
  end
  
end

class Gameserver

  include Constant::Object
  include Constant::Gameserver
  include Constant::Cavelevel
  include Scroll
  include Potion
  include Display
  include Utility
  include Wizard
  include Store
  include Trap
  include Command
  include Regenerate
  include Movemonster
  include Spell
  include Death
  
  # server:   ᡣԡGameserver٤Ρ
  # level:    
  # hardgame: ١

  START = 0

  attr_reader :cavelevel
  
  def initialize( server, database, level, load_data = nil )

    @server   = server    
    @database = database  
    @level    = level      
    
    # 󥸥
    @cavelevel = ( load_data ? load_data : Cavelevel.new( level ) )

    # intelligent_moveѡ
    @intelligent_map = Distancemap.new
    @vampire_map     = Distancemap.new

    # 1󤴤Ȥ˰Ĥͤ򸺤餷0ʲˤʤȥ󥹥֤ޤ
    @rmst     = 80
    
  end

  def player2screen( player )
    player.screen[ @level ]
  end
  
  def town?
    @level == START
  end
  
=begin
Spells:  1( 1)  AC: 0    WC: 0    Level 1  Exp: 0           novice explorer  
HP:  10( 10) STR=12 INT=12 WIS=12 CON=12 DEX=12 CHA=12 LV: H  Gold: 0     
=end
  # 59 + ιĹ
  SFORMAT1 = "Spells:%3d(%3d)  AC: %-3d  WC: %-3d  Level%3d Exp: %-9d %s"
  # 72 + ̾
  SFORMAT2 = "HP: %3d(%3d) STR=%-2d INT=%-2d WIS=%-2d CON=%-2d DEX=%-2d CHA=%-2d LV: %s  Gold: %-6d"     
  
  def print_status( client )

    player = client.player
    mesg   = player.mesg
    
    client.status1( mesg[ 1540 ], player.status1 )
    client.status2( mesg[ 1541 ], player.status2( @level ) )
        
  end

  def message( client, str, *arg )
    client.message( sprintf( str, *arg ) )
  end
  
  # ϲµܤ鵢ԡ
  def from_underground( client ) 

    raise unless @level == START

    join( client, @cavelevel.entrance )

  end

  def enter( client )

    raise unless @level == 1
    
    join( client, @cavelevel.exit )
        
  end

  # 볦롣
  def create_screen( player )

    unless @cavelevel.already_joined?( player )
      player.screen[ @level ] = Screen.new
    end

  end

  def volcano?
    @level >= MAXLEVEL 
  end
  
  def underground?
    ( @level != 0 ) && !volcano?
  end

  def teleport_join( client )
    join( client )
  end

  def pit_join( client, point )
    join( client, point, :pit_damage )
  end

  def fall_join( client, point )
    join( client, point, :fall_damage )
  end

  def climb_down( client )
    join( client, @cavelevel.shaft_up, :climb_down_damage )
  end

  def climb_up( client )
    join( client, @cavelevel.shaft_down )
  end

  # ץ쥤䡼ƤӤγˤäƤȤˤϡ
  # 󥹥֤ޤ
  
  def fill_monsters_reenter( client )
    @cavelevel.fill_monsters_reenter( client.player )
  end
  
  def join( client, point = nil, join_damage = nil )

    player = client.player

    player.reset_teleflag if @level == 0

    @cavelevel.wipeout_monsters
    fill_monsters_reenter( client )
    create_screen( player )
    client.newcavelevel( player2screen( player ), @cavelevel.volcano? )
    
    # ޤǤ˼̤ƤϷʤɤɽ
    showall( client ) if @level == 0
    drawscreen( client )
  
    # ץ쥤䡼֤롣
    put_player( client, point )
    @cavelevel.join( player )

    # Ȥ꤫Ƥʤɤˡ᡼Ϳ롣
    send( join_damage, client ) if join_damage
    
  end

  def load( client )
    return unless client.player = @cavelevel.name2player( client.name )
    client.newcavelevel( player2screen( client.player ), @cavelevel.volcano? )
    drawscreen( client )
  end
  
  DIR = "hjklyubn"

  def random_dir
    DIR[ rand( DIR.size ) ]
  end

  def confuse_move?( player ) # Ļ­
    ( player.confuse? && ( player.level < rnd( 30 ) ) )
  end
  
  def move_player( client, from, dir )

    player = client.player
    
    to   = next_point( from, ( confuse_move?( player ) ) ? random_dir : dir )

    if ( monst = @cavelevel.monsters[ to ] ) # ư˥󥹥鹶⡣
      hitmonster( client, monst, to )
      true # ưλ
    elsif !( @cavelevel.movable?( player, to ) )
      return false # ưľ
    else
      do_move_player( client, from, to )
    end
    
  end

  def do_move_player( client, from, to )
    @cavelevel.move_player( client.player, from, to )
    showcell( client, to )
    lookforobject( client, from, to )
  end
  
  def oexit( client )
    leave( client )
    @server.push( [ :from_underground, client ] )
  end

  def add_monster( client )
    
    return if client.player.timestop?

    if ( @rmst -= 1 ) <= 0 # 󥿡Ĥĸ餹
      @rmst = ( 120 - ( @level << 2 ) )
      return unless ( point = @cavelevel.add_monster ) # ֤Ǥ
      update_point( client, point )
    end

  end

  
  # ٥ȥ롼ס
  def main( client )

    player = client.player
    
    loop do
      
      break unless @cavelevel.player_exist?( player )
    
      srand

      client.time_left( player.time_left )
      print_status( client )
      side_status( client )
            
      move_all_monsters( client ) # player̤⡣
      break unless @cavelevel.player_exist?( player )
      showcell( client, @cavelevel.player2point( player ) )

      move_all_monsters( client ) if player.hastemonst?
      move_all_spheres( client )
      break unless @cavelevel.player_exist?( player )
      
      showcell( client, @cavelevel.player2point( player ) )
      
      print_status( client )
      side_status( client )

      command( client ) if player.movable?

      break if @server.find { | req | req && ( req[ 0 ] == :save ) }
      break unless @cavelevel.player_exist?( player )
      
      turn_end( client )
      add_monster( client )
     
      showcell( client, @cavelevel.player2point( player ) )
      
    end

  end

  def turn_end( client )

    player = client.player
    
    player.time_passed
    player.interest

    regenerate( client )
    
  end
  
  def version( client )

    client.version( client.player.mesg[ 0 .. 4 ].collect do | string |
		     sprintf( string, $HARDGAME )
		   end )
    
    #  if (wizard) lprcat(" Wizard"); nomove=1;
    # if (cheat) lprcat(" Cheater");
    # message( client, $LARN_COPYRIGHT )
    
  end

  def itemlist_without_time( client, list )
    client.itemlist( list, client.player.mesg[ 152 .. 154 ] ) 
  end
  
  def equip_list( client )
    itemlist_without_time( client, client.player.equip_list )
  end

  def list_ring( client )
    itemlist_without_time( client, client.player.list_ring )
  end
  
  def all_item( client )

    player = client.player
    mesg   = player.mesg

    client.itemlist( player.all_item, mesg[ 152 .. 154 ],
		    sprintf( mesg[ 151 ], player.mobuls_elapsed,
			    player.mobuls_left ) )
    
  end
  
  def command( client )

    loop do
    
      point = @cavelevel.player2point( client.player )
    
      case ( com = client.command )
      when ?h, ?j, ?k, ?l, ?y, ?u, ?b, ?n
	move_player( client, point, com )
	return 
      when ?v
	version( client )
      when ?Z
	teleport_as_command( client )
	return 
      when ?i
	all_item( client )
      when ?]
	equip_list( client )
      when ?=
	list_ring( client )
      when ?^
	identify_trap( client, point )
      when ?e
	pocket_eat( client )
	return 
      when ?r
	pocket_read( client, point )
	return
      when ?q
	pocket_quaff( client, point )
	return
      when ?c
	cast_spell( client, point )
	return
      when ??
	help( client )
      when ?W
	wear( client )
	return
      when ?w
	wield( client )
	return
      when ?T
	take_off( client )
	return
      when ?.
	stay_here( client )
	return
      when ?>
	command_go_down( client, point )
	return
      when ?<
	command_go_up( client, point )
	return
      when ?d
	drop( client, point )
	return
      when ?g
	pack_weight( client )
      when ?/, ?x
	investigate( client )
      when ?P
	tax_status( client )
      when ?D, ?\\
	all_item_found( client )
      when ?S
	save_game( client )
	return
      when ?Q
	quit_game( client )
	return
      when ?C
	drop_chest( client, point )
      when ?R
	drop_scroll( client, point )
      when ?\M-q
	drop_potion( client, point )
      when ?A
	drop_altar( client, point )
      when ?\M-m
	drop_mirror( client, point )
      when ?\M-t
	drop_throne( client, point )
      when ?X
	charge_tax( client )
      when ?G
	enough_gold( client )
      when ?M
	debug_put_monster( client, point )
      when ?\cd
	debug_put_closeddoor( client, point )
      when ?\M-s
	debug_spell_learn_id( client )
      when ?\M-l
	debug_raise_level( client )
      when ?\M-^
	debug_all_traps_found( client, point )
      when ?\M-k
	debug_kill_100_mobuls( client )
      when ?\M-g
	debug_drop_gem( client, point )
      when ?\M-d
	debug_identify_all( client )
      else
	
      end

    end
      
  end

  def save_game( client )
    message( client, client.player.mesg[ 1570 ] )
    @server.push( [ :save, client ] )
  end

  def quit_game( client )

    mesg = client.player.mesg

    message( client, mesg[ 1575 ] )

    answer = client_quit_question( client, mesg[ 1576 ],
				  { ?y => mesg[ 1577 ], ?Y => mesg[ 1577 ],
				    ?n => mesg[ 1578 ], ?N => mesg[ 1578 ],
				    ?\e => mesg[ 1578 ] },
				  mesg[ 1579 .. 1582 ] )

    if ( answer == ?y ) || ( answer == ?Y )
      exit
    end
    
  end

  def client_quit_question( client, question, answer, second_question ) 
    client.quit_question( question, answer, second_question )
  end
  
  
  def put_player( client, point )

    point = @cavelevel.put_player( client.player, point )
    showcell( client, point )
    
  end

  def ogold( client, object, point )
    
    player = client.player
    
    # 򽦤äȤΡ
    client.gold( player.mesg[ 256 .. 257 ], amount = object.amount ) 

    client.player.get_gold( amount ) # ۤˡ
    lose_object( client, object, point )

    print_status( client )           # ⤬ΤǡơԽ񤭽Ф

  end

  def lose_object( client, object, point )
    @cavelevel.lose_object( object ) # 󥸥⤫õ
    showcell_radius( client, point, 0 )        # 衣
  end

  def get_object( client, object, point )
    @cavelevel.get_object( object, point )
    showcell_radius( client, point, 0 )
  end


  def oaltar( client, object, point )
    
    player = client.player
    mesg   = player.mesg

    return if @cavelevel.nearbymonst?( player )

    answer = reach_question( client, mesg[ 1421 ], mesg[ 1422 ],
			    { ?p => mesg[ 1423 ],
			      ?d => mesg[ 1424 ],
			      ?i => mesg[ 1425 ], ?\e => mesg[ 1425 ] } )

    case answer
    when ?i, ?\e
      altar_ignore( client, object, point ) # ̵뤷°Ĥ
    when ?p
      altar_pray( client, object, point )   # 롣
    when ?d
      altar_desecrate( client, object, point ) # Ť¡
    end

  end

  def altar_ignore( client, object, point )
    
    player = client.player
    
    if ( rnd( 100 ) < 30 )
      put_monster_near( client, Monster.create( @level + 1 ), point )
      player.get_aggravate( rnd( 450 ) )
    else
      message( client, player.mesg[ 1426 ] )
    end
    
  end

  def altar_desecrate( client, object, point ) # ˷ޤޤ

    player = client.player
    mesg   = player.mesg
    
    if ( rnd( 100 ) < 60 )
      put_monster_near( client, Monster.create( @level + 2, 8 ), point )
      player.get_aggravate( 2500 )
    elsif ( rnd( 101 ) < 30 ) # žǡ
      message( client, mesg[ 1427 ] )
      lose_object( client, object, point )
    else
      message( client, mesg[ 1428 ] )  # ʤ
    end

  end

  def altar_hear( client )

    player = client.player
    
    message( client, player.mesg[ 1434 ] ) # ɸϥåס
    player.get_altpro

  end

  def altar_enchant_armor( client )

    player = client.player

    ac = player.ac
    player.enchant_armor
    message( client, player.mesg[ 1435 ] ) if player.ac != ac
    
  end

  def altar_enchant_weapon( client )

    player = client.player

    if player.wield
      message( client, player.mesg[ 1436 ] )
      player.enchant_weapon
    end
    
  end
  
  def altar_just_pray( client, object, point )

    player = client.player
    mesg   = player.mesg

    if ( rnd( 100 ) < 75 )
      message( client, mesg[ 1433 ] ) # ⵯʤ
    elsif ( rnd( 13 ) < 4 )
      altar_hear( client )
    elsif ( rnd( 43 ) == 10 ) # ɸϥåס
      altar_enchant_armor( client )
    elsif ( rnd( 43 ) == 10 ) # ΰϥåס
      altar_enchant_weapon( client )
    else
      put_monster_near( client, Monster.create( @level + 1 ), point )
    end

  end

  def client_altar_get_amount( client )
    client.altar_get_amount( client.player.mesg[ 1437 ] )
  end

  def altar_pray_with_money( client, object, point )

    player = client.player
    mesg   = player.mesg

    if ( amount = client_altar_get_amount( client ) ) > player.gold
      message( client, mesg[ 1438 ] )
      return
    end

    player.lose_gold( amount )

    # ⤬ʤܤ餷ʿ͡
    if ( amount < player.gold / 10 ) || ( amount < rnd( 50 ) )
      put_monster_near( client, Monster.create( @level + 1 ), point )
      player.get_aggravate( 200 )
    elsif ( rnd( 101 ) > 50 )
      altar_hear( client )
    elsif ( rnd( 43 ) == 5 )
      altar_enchant_armor( client )
    elsif ( rnd( 43 ) == 8 )
      altar_enchant_weapon( client )
    else
      message( client, mesg[ 1439 ] ) # ʧäʤ
    end
      
  end
  
  def altar_pray( client, object, point )   # 롣

    player = client.player
    mesg   = player.mesg

    answer = question( client, mesg[ 1429 ],
		      { ?j => mesg[ 1431 ],
			?m => mesg[ 1430 ], ?\e => mesg[ 1432 ] } )

    case answer
    when ?j
      altar_just_pray( client, object, point )
    when ?m
      altar_pray_with_money( client, object, point )
    when ?\e
      return
    end

  end
  
  NFORTUNES = 34

  def eat( client, object, point )

    player = client.player
    mesg   = player.mesg
    
    if player.blindcount?
      message( client, mesg[ 45 ] )
    else
      message( client, mesg[ 46 ] )
      message( client, mesg[ 5 + rand( NFORTUNES ) ] ) # fortune
    end

    lose_object( client, object, point ) # cookie볦⤫õ
    
  end

  def item( client, object )
    client.item( client.player.item( object ) )
  end
  
  def take( client, object, point )

    player = client.player
    mesg   = player.mesg
    
    if player.pocketfull?
      message( client, mesg[ 48 ] )
      return
    end

    message( client, mesg[ 47 ] )
    player.pack_item( object )
    item( client, object )

    lose_object( client, object, point ) # ƥ볦⤫
      
  end

  
  def ocookie( client, object, point )

    mesg = client.player.mesg

    answer = reach_question( client, mesg[ 40 ], mesg[ 41 ],
			    { ?\e => mesg[ 42 ], ?i  => mesg[ 42 ],
			      ?e  => mesg[ 43 ],
			      ?t  => mesg[ 44 ] } ) 
    case answer
    when ?\e, ?i
      return
    when ?e
      eat( client, object, point )
    when ?t
      take( client, object, point )
    end

  end

  def pry_off( client, object, point )

    if ( k = rnd( 101 ) ) < 25 
      rnd( 4 ).times { creategem( client, point ) }
      lose_object( client, object, point )
      get_object(  client, Larnobject.deadthrone, point )
    elsif ( k < 40 ) && object.throne?
      put_monster_near( client, Monster.gnomeking, point )
      lose_object( client, object, point )
      get_object( client, Larnobject.throne2, point )
    else
      message( client, client.player.mesg[ 371 ] )
    end

  end

  def creategem( client, point )

    case rnd( 4 )
    when 1
      gem = Larnobject.new( ODIAMOND,  rnd( 50 ) + 50 / 10 )
    when 2
      gem = Larnobject.new( ORUBY,     rnd( 40 ) + 40 / 10 )
    when 3
      gem = Larnobject.new( OEMERALD,  rnd( 30 ) + 30 / 10 )
    when 4
      gem = Larnobject.new( OSAPPHIRE, rnd( 20 ) + 20 / 10 )
    end

    p_point = @cavelevel.put_object_near( gem, point )
    showcell_radius( client, p_point, 1 ) if p_point

  end
  
  def put_monster_near( client, monst, point )
    p_point = @cavelevel.put_monster_near( monst, point )
    client_movemonst( client, monst, nil, p_point ) if p_point
  end

  def put_monster( client, monster, point )

    return false unless @cavelevel.put_monster( monster, point ) # Բġ
    update_point( client, point )

  end
  
  def sit_down( client, object, point )

    mesg = client.player.mesg

    if ( k = rnd( 101 ) ) < 30 && object.throne?
      put_monster_near( client, Monster.gnomeking, point )
      lose_object( client, object, point )
      get_object( client, Larnobject.throne2, point )
    elsif
      message( client, mesg[ 372 ] )
      teleport( client )
    else
      message( client, mesg[ 373 ] )
    end
      
  end
    
  def othrone( client, object, point )

    mesg = client.player.mesg

    answer = reach_question( client, mesg[ object.throne? ? 366 : 375 ],
			    mesg[ 367 ],
			    { ?p => mesg[ 368 ],
			      ?s => mesg[ 369 ],
			      ?i => mesg[ 370 ], ?\e => mesg[ 370 ] } )
    case answer
    when ?i, ?\e
      return
    when ?p
      pry_off( client, object, point )
    when ?s
      sit_down( client, object, point )
    end

  end

  def odeadthrone( client, object )

    mesg = client.player.mesg

    answer = reach_question( client, mesg[ 351 ], mesg[ 352 ],
			    { ?s => mesg[ 353 ],
			      ?i => mesg[ 354 ], ?\e => mesg[ 354 ] } )

    case answer
    when ?i, ?\e
      return
    when ?s
      if rnd( 101 ) < 35
	message( client, mesg[ 355 ] )
	teleport( client )
      else
	message( client, mesg[ 356 ] )
      end
    end

  end

  def teleport( client, error = false )

    if error && ( rnd( 151 ) < 3 )
      death_solid_rock( client )
    end

    # ƥݡȤŪϡ
    # 鳹ءϲµܤϲµܤء

    client.player.teleport

    if @level == 0
      dest = 0
    elsif underground?
      dest = rnd( 5 ) + @level - 3
      dest = 1 if dest < 1
      dest = MAXLEVEL - 1 if dest >= MAXLEVEL
    else
      dest = rnd( 3 ) + @level - 2
      dest = MAXLEVEL if dest < MAXLEVEL
      dest = MAXLEVEL + MAXVLEVEL - 1 if dest >= MAXLEVEL + MAXVLEVEL
    end

    leave( client )
    @server.push( [ :teleport_join, client, dest ] )
    
  end

  def oorb( client, object )

    player = client.player
    mesg   = player.mesg

    mesg[ 395 .. 396 ].each do | str |
      message( client, str )
    end
    
  end

  def do_fall_pit( client, object, point )

    player = client.player
    mesg   = player.mesg
    
    # ʤܤ¨ࡣ
    if @cavelevel.bottom?
      message( client, mesg[ 404 ] )
      death_bottomless_pit( client )
    else
      leave( client )
      @server.push( [ :pit_join, client, @level + 1, point ] )
    end

  end
  
  def opit( client, object, point )

    player = client.player
    mesg   = player.mesg

    client.pit_message( mesg[ 401 ] )
   
    return unless player.pit_fall?

    do_fall_pit( client, object, point )

  end
  
  def pit_damage( client )

    player = client.player
    mesg   = player.mesg
    
    if ( rnd( 101 ) < 20 )
      message( client, mesg[ 402 ] )
    else
      damage = rnd( @level * 3 + 3 )
      message( client, mesg[ 403 ], damage )
      death_pit( client, damage )
    end
      
  end

  def kick( client )

    mesg = client.player.mesg
    
    if ( rnd( 2 ) == 1 )
      message( client, mesg[ 1848 ] )
    else
      damage = rnd( ( @level + 1 ) << 1 )
      message( client, mesg[ 1849 ], damage )
      death_stupid_frustration( client, damage )
    end

  end

  def do_go_up( client, point )

    leave( client )
    @server.push( [ :move, client, @level - 1, point ] )

  end

  
  def go_up( client, object, point )

    if object.stairsdown?
      message( client, client.player.mesg[ 1850 ] )
      return
    end

    do_go_up( client, point )
    
  end

  def do_go_down( client, point )

    leave( client )
    @server.push( [ :move, client, @level + 1, point ] )

  end
  
  def go_down( client, object, point )

    if object.stairsup?
      message( client, client.player.mesg[ 1852 ] )
      return
    end
    
    do_go_down( client, point )

  end

  def ostairs( client, object, point )

    mesg = client.player.mesg

    answer = reach_question( client, mesg[ 1841 ],
		      ( object.stairsdown? ? mesg[ 1843 ] : mesg[ 1842 ] ),
		      {	?s  => mesg[ 1844 ],
			?\e => mesg[ 1844 ],
			?i  => mesg[ 1844 ],
			?u  => mesg[ 1845 ],
			?d  => mesg[ 1846 ],
			?f  => mesg[ 1847 ],
			?k  => mesg[ 1847 ], } )
    
    case answer
    when ?f, ?k
      kick( client )
    when ?u
      go_up( client, object, point )
    when ?d
      go_down( client, object, point )
    end
    
  end

  def drink( client, object, point )

    player = client.player
    mesg   = player.mesg
        
    if ( rnd( 1501 ) < 2 )
      message( client, mesg[ 386 ] )
      death_dreadful_sleep( client )
      return
    end

    if ( x = rnd( 100 ) ) < 7
      player.get_halfdam
      message( client, mesg[ 387 ] )
    elsif ( x < 13 )
      # quaffpotion(23); /* see invisible */
    elsif ( x < 45 )
      message( client, mesg[ 388 ] )
    elsif ( rnd( 3 ) !=  2 )
      fntchange( client, 1 )
    else
      fntchange( client, -1 )
      
      if ( rnd( 12 ) < 3 )
	message( client, mesg[ 389 ] )
	lose_object( client, object, point )
	get_object( client, Larnobject.deadfountain, point )
      end

    end
    
  end

  def fntchange( client, how )

    player = client.player
    mesg   = player.mesg
    
    case rnd( 9 )
    when 1

      message( client, mesg[ ( how < 0 )? 410 : 411 ] )
      player.strength_up( how )

    when 2

      message( client, mesg[ ( how < 0 )? 412 : 413 ] )
      player.intelligence_up( how )

    when 3

      message( client, mesg[ ( how < 0 )? 414 : 415 ] )
      player.wisdom_up( how )

    when 4

      message( client, mesg[ ( how < 0 )? 416 : 417 ] )
      player.constitution_up( how )

    when 5

      message( client, mesg[ ( how < 0 )? 418 : 419  ] )
      player.dexterity_up( how )

    when 6
      
      message( client, mesg[ ( how < 0 )? 420 : 421  ] )
      player.charisma_up( how )

    when 7
      
      diff = rnd( @level + 1 )
      
      if how < 0
	message( client, mesg[ ( diff == 1 )? 422 : 423 ], diff )
	player.losemhp( diff )
      else
	message( client, mesg[ ( diff == 1 )? 424 : 425 ], diff )
	player.raisemhp( diff )
      end

    when 8

      diff = rnd( @level + 1 )
      
      if ( how < 0 )
	message( client, mesg[ ( diff == 1 )? 426 : 427 ], diff )
	player.losemspell( diff )
      else
	message( client, mesg[ ( diff == 1 )? 428 : 429 ], diff )
	player.raisemspell( diff )
      end

    when 9

      diff = 5 * rnd( ( @level + 1 ) * ( @level + 1 ) )
      
      if ( how < 0 )
	message( client, mesg[ ( diff == 1 )? 430 : 431 ], diff )
	loseexperience( client, diff )
      else
	message( client, mesg[ ( diff == 1 )? 432 : 433 ], diff )
	raiseexperience( client, diff )
      end

    end
    
  end

  def loseexperience( client, diff )

    player = client.player
    mesg   = player.mesg

    prelevel = player.level
    player.loseexperience( diff )

    if ( prelevel != player.level )
      # client.beep
      message( client, mesg[ 435 ], player.level )
    end
    
  end

  def raiseexperience( client, diff )

    player = client.player
    mesg   = player.mesg

    prelevel = player.level
    player.raiseexperience( diff )

    if ( prelevel != player.level )
      # client.beep
      message( client, mesg[ 436 ], player.level )
    end

  end
  
  def wash_yourself( client, object, point )

    player = client.player
    mesg   = player.mesg
    
    if ( rnd( 100 ) < 11 )
      damage = rnd( ( @level << 2 ) + 2 )
      message( client, mesg[ 390 ], damage )
      death_drank_poison( client, damage )
    elsif ( rnd( 100 ) < 29 )
      message( client, mesg[ 391 ] )
    elsif ( rnd( 100 ) < 31 )
      message( client, mesg[ 392 ] )
    elsif ( rnd( 100 ) < 34 )
      put_monster_near( client, Monster.waterlord, point )
    else

      message( client, mesg[ 393 ] )

    end

  end

  
  def ofountain( client, object, point )

    return if @cavelevel.nearbymonst?( client.player )

    mesg = client.player.mesg

    answer = reach_question( client, mesg[ 381 ], mesg[ 382 ],
			    { ?d => mesg[ 383 ],
			      ?w => mesg[ 384 ],
			      ?i => mesg[ 385 ], ?\e => mesg[ 385 ] } )
    
    case answer
    when ?i, ?\e
      return
    when ?d
      drink( client, object, point )
    when ?w
      wash_yourself( client, object, point )
    end

  end

  def ostatue( client, object )
    
    return if @cavelevel.nearbymonst?( client.player )

    client.statue_message( client.player.mesg[ 261 ] )
      
  end

  def open_chest( client, object, point )

    player = client.player
    mesg   = player.mesg

    unless rnd( 101 ) < 40
      message( client, mesg[ 641 ] ) # Ȣϳ
      return
    end

    # ȢƱȯ
    message( client, mesg[ 636 ] )
    message( client, mesg[ 637 ], damage = rnd( 10 ) )
    death_exploding_chest( client, damage )

    # Ȣμ
    case ( rnd( 10 ) )
    when 1 # ߡ
      message( client, mesg[ 638 ] )
      player.get_itching
    when 2 # ԴѤ郎ޤʤʤޤ
      message( client, mesg[ 639 ] )
      player.get_clumsiness 
    when 3 # ᡼Ⱦ
      message( client, mesg[ 640 ] )
      player.get_terrible_halfdam
    end
    
    lose_object( client, object, point ) # Ȣü
    
    # ФȤ⡣
    creategem( client, point ) if ( rnd( 101 ) < 69 )
    dropgold_amount( client, 110 * object.arg + 200, point )
    
    # ƥФޤ
    rnd( 4 ).times { something( client, point, object.arg + 2 ) }
    
  end
  
  def ochest( client, object, point )

    player = client.player
    mesg   = player.mesg

    answer = reach_question( client, mesg[ 631 ], mesg[ 632 ],
			    { ?t => mesg[ 633 ],
			      ?o => mesg[ 634 ],
			      ?i => mesg[ 635 ], ?\e => mesg[ 635 ] } )

    case answer
    when ?i, ?\e
      return
    when ?o
      open_chest( client, object, point )
    when ?t
      take( client, object, point )
    end

  end
  
  def omirror( client, object )

    return if @cavelevel.nearbymonst?( client.player )

    client.mirror_message( client.player.mesg[ 266 ] )

  end

  def odeadfountain( client, object )

    return if @cavelevel.nearbymonst?( client.player )

    client.player.mesg[ 300 .. 301 ].each do | str |
      message( client, str )
    end

  end

  def close_door( client, object, from, to )

    # ɥĤơ
    object.closed
    update_point( client, to )

    # ξء
    do_move_player( client, to, from )
    
  end
    
  def oopendoor( client, object, from, to )

    player = client.player
    mesg   = player.mesg
    
    answer = reach_question( client, mesg[ 1301 ], mesg[ 1302 ],
			    { ?c => mesg[ 1303 ],
			      ?i => mesg[ 1304 ], ?\e => mesg[ 1304 ] } )

    close_door( client, object, from, to ) if answer == ?c
    
  end

  def open_door( client, door, from, to )

    player = client.player
    mesg   = player.mesg

    if player.open_success?
    
      # ɥ櫤ųݤƤ뤳Ȥޤա
      # Larnξ硢ɥϳΤǤϤʤ
      # ˡǤ֤ΤǤ
      
      if door.door_aggravate?
     	player.get_aggravate( rnd( 400 ) )
      elsif door.door_electric? # ŵå
	message( client, mesg[ 1310 ] )
	death_electric_shock( client, rnd( 20 ) )
      elsif door.door_drainlevel? # ٥㲼
	do_drain_level( client )
      elsif door.door_strength_down? # 㲼
	message( client, mesg[ 1311 ] )
	player.strength_down
      end
      
      door.opened
      update_point( client, to )

    else
      # ɥʤäΤǡξء
      do_move_player( client, to, from )

    end

  end

  def oclosedoor( client, object, from, to )

    player = client.player
    mesg   = player.mesg

    answer = reach_question( client, mesg[ 1306 ], mesg[ 1307 ],
		      { ?o => mesg[ 1308 ],
			?i => mesg[ 1309 ], ?\e => mesg[ 1309 ] } )

    case answer
    when ?i, ?\e
      do_move_player( client, to, from )
    when ?o
      open_door( client, object, from, to )
    end

  end


  def leave( client )
    
    player = client.player
    point  = @cavelevel.player2point( player )
    
    @cavelevel.leave( player )
    showcell( client, point )

  end

  
  def oentrance( client, object )

    mesg = client.player.mesg

    answer = reach_question( client, mesg[ 360 ], mesg[ 361 ],
			    { ?g => mesg[ 362 ],
			      ?i => mesg[ 363 ], ?\e => mesg[ 363 ] } )
    
    return if answer != ?g

    # ϲµܤΰ쳬ء
    
    leave( client )
    @server.push( [ :enter, client ] )
    
  end

  def ovoldown( client, object, point )

    player = client.player
    mesg   = player.mesg

    answer = reach_question( client, mesg[ 1440 ], mesg[ 1441 ],
			    { ?c => mesg[ 1442 ],
			      ?i => mesg[ 1443 ], ?\e => mesg[ 1443 ] } )

    return if answer != ?c
    
    leave( client )
    @server.push( [ :climb_down, client ] )

  end

  def climb_down_damage( client )

    player = client.player
    
    if player.climb_down_damage?
      message( client, player.mesg[ 1445 ] )
      death_slipped_shaft( client, 30 + rnd( 20 ) )
    end
    
  end
  
  def ovolup( client, object, point )

    player = client.player
    mesg   = player.mesg

    answer = reach_question( client, mesg[ 1450 ], mesg[ 1451 ],
			    { ?c => mesg[ 1452 ],
			      ?i => mesg[ 1453 ], ?\e => mesg[ 1453 ] } )

    return if answer != ?c
        
    if player.climb_up_damage? # ʪŤơФʤ
      message( client, mesg[ 1455 ] )
      death_slipped_shaft( client, 15 + rnd( 20 ) )
      return
    end
    
    leave( client ) # γϽλ
    @server.push( [ :climb_up, client ] ) # ء

  end

  def question( client, str, choice )
    client.question( str, choice )
  end

  def lookforobject( client, from, point )

    # /* can't find objects is time is stopped */
    return if ( player = client.player ).timestop?

    if @cavelevel.point2sphere( point )
      death_sphere_of_annihilation( client )
      return
    end
   
    return unless ( object = @cavelevel.point2object( point ) )

    if object.gold?
      ogold( client, object, point )
    elsif object.potion?
      opotion( client, object, point )
    elsif object.altar?
      oaltar( client, object, point )
    elsif object.book?
      obook( client, object, point )
    elsif object.scroll?
      oscroll( client, object, point )
    elsif object.cookie?
      ocookie( client, object, point )
    elsif object.throne? || object.throne2?
      othrone( client, object, point )
    elsif object.deadthrone?
      odeadthrone( client, object )
    elsif object.orb?
      oorb( client, object )
    elsif object.pit?
      opit( client, object, point )
    elsif object.stairsup? || object.stairsdown?
      ostairs( client, object, point )
    elsif object.elevatorup? ||  object.elevatordown?
     # Larn12.0Ǥϡ̤
    elsif object.fountain?
      ofountain( client, object, point )
    elsif object.statue?
      ostatue( client, object )
    elsif object.chest?
      ochest( client, object, point )
    elsif object.trap?
      otrap( client, object, point )
    elsif object.store?
      ostore( client, object, point )
    elsif object.mirror?
      omirror( client, object )
    elsif object.deadfountain?
      odeadfountain( client, object )
    elsif object.opendoor?
      oopendoor( client, object, from, point )
    elsif object.closeddoor?
      oclosedoor( client, object, from, point )
    elsif object.entrance?
      oentrance( client, object )
    elsif object.voldown?
      ovoldown( client, object, point )
    elsif object.volup?
      ovolup( client, object, point )
    elsif object.exit?
      oexit( client )
    else
      finditem( client, object, point )
    end
    
  end

  def finditem( client, object, point )

    player = client.player
    mesg   = player.mesg
    ench   = object.enchant
    
    str = if ench == 0
	    sprintf( mesg[ 526 ], player.objectname( object ) )
	  elsif ench > 0
	    sprintf( mesg[ 527 ], player.objectname( object ), ench )
	  elsif ench < 0
	    sprintf( mesg[ 528 ], player.objectname( object ), ench )
    end

    answer = reach_question( client, str, mesg[ 529 ],
			    { ?t => mesg[ 530 ],
			      ?i => mesg[ 531 ], ?\e => mesg[ 531 ] } )

    case answer
    when ?i, ?\e
      return
    when ?t
      take( client, object, point )
    end
    
  end

  def pocket_eat( client )

    player = client.player
    mesg   = player.mesg
  
    choice = client.select( mesg[ 250 ], player.list_eatable,
			   mesg[ 152 .. 154 ], mesg [ 251 ] )

    return if choice == ?\e

    cookie = player.pocket( choice )

    if !cookie
      message( client, mesg[ 254 ], choice )
      return true
    end

    if !cookie.cookie?
      message( client, mesg[ 255 ] )
      return true
    end

    message( client, mesg[ 5 + rand( NFORTUNES ) ] ) # fortune
    player.lose_item( cookie )

  end

  def get_direction( client )
    client.get_direction( client.player.mesg[ 665 ] )
  end
  
  # ˡ򾧤ȤʤѤݤ˺𤷤Ƥȡ
  # ޤդޤ
  
  def isconfuse?( client )

    player = client.player
    mesg   = player.mesg
    
    if player.confuse?
      message( client, mesg[ 664 ] )
      true
    else
      false
    end

  end

  # from: ˡȯꡣӾԤξꡣ
  # dam:  ˡΰϡ
    
  def go_direct( client, from, spell, dam, mesg )

    player = client.player
 
    return if isconfuse?( client ) # 𤷤ƤΤǡդ

    recursive_go_direct( client, from, spell, dam, mesg,
			get_direction( client ) )
		

  end

  # monsterФơspell̵ʤȤˤtrue
  def nospell?( client, monster, spell )

    player = client.player
    mesg   = player.mesg
    
    if  !( anti_spell = monster.resist_spell?( spell ) )
      return false
    else
      message( client, mesg[ 680 + anti_spell ],
	      player.real_monstername( monster ) )
      return true
    end
	      
  end

  def crush_door( client, object, point, damage, mesg = nil )

    message( client, mesg[ 0 ] ) if mesg
    
    if ( damage >= 40 )
      message( client, mesg[ 1 ] ) if mesg
      lose_object( client, object, point )
    end
    
  end

  def crush_statue( client, object, point, damage, mesg = nil )

    message( client, mesg[ 0 ] ) if mesg

    if ( $HARDGAME < 3 ) && ( damage > 44 )
      message( client, mesg[ 1 ] ) if mesg
      lose_object( client, object, point ) # Ħ...
      get_object( client, Larnobject.newbook( @level ), point ) # 椫ˡ񤬡
    end

  end

  def throne_attack( client, object, point, mesg = nil )

    message( client, mesg ) if mesg

    @cavelevel.put_monster( Monster.gnomeking, point )
    lose_object( client, object, point )
    get_object( client, Larnobject.throne2, point )
    
  end
  
  def client_magic_attack( client, spell, point )
    client.magic_attack( spell, point )
  end
			  
  def recursive_go_direct( client, from, spell, dam, mesg, dir )

    player = client.player
    
    return unless dam > 0 # ˡ⤳ޤǡ

    to = next_point( from, dir )
    return if @cavelevel.out_of_range?( to ) # 󥸥γء

    client_magic_attack( client, spell, to )
    
    showcell_radius( client, to, 0 )
    
    if ( another = @cavelevel.point2player( to ) ) # ץ쥤䡼˥ҥåȡ
      
      message( client, mesg[ 0 ] )
      # client = player2client( another )
      death_own_magic( client, dam )

    elsif ( monster = @cavelevel.point2monster( to ) ) # 󥹥˥ҥåȡ
      
      return if nospell?( client, monster, spell ) # ʸ̵ʤ...

      message( client, mesg[ 1 ], player.monstername( monster ) )

      dam -= monsterdamaged( client, monster, to,
			    player.spell_damage( monster, dam ) )
      
      return unless monster.killed? # 󥹥ʤʤä餽ޤǡ

    elsif ( @cavelevel.wall?( to ) )

      message( client, mesg[ 2 ] )

      if @cavelevel.wall_crumbles?( to, dam )
	message( client, mesg[ 3 ] )
	@cavelevel.wall_crumbles( to )
	showcell_radius( client, to, 0 )
      end
      
      return # ɤä顢Ƶ⤳ޤǡ
      
    elsif ( object = @cavelevel.point2object( to ) )

      if object.closeddoor?
	crush_door( client, object, to, dam, mesg[ 4 .. 5 ] )
	return
      elsif object.statue?
	crush_statue( client, object, to, dam, mesg[ 6 .. 7 ] )
	return
      elsif object.throne?
	throne_attack( client, object, to, mesg[ 8 ] )
	return
      elsif object.mirror?
	dir = opposite_dir( dir )
      end

    end
    
    dam -= ( 3 + ( $HARDGAME >> 1 ) )
    
    recursive_go_direct( client, to, spell, dam, mesg, dir )

  end

  # 󥹥monstfromtoذưȤ򡢥饤ȤΤ롣
  # 󥹥Ȥˤϡfromnil
  # 󥹥˴Ȥˤϡtonil
  
  def client_movemonst( client, monst, from, to )

    # fromtoΤ줾ˤĤΤԤ
    # ͩͩˤϡ̾Υ󥹥ȤΤ롣
    # ͩʤȤˤϡƥäȤ¸ߤΤ롣
    # ǡˤĤƤƱ͡

    update_point_no_update( client, from ) 
    update_point_no_update( client, to   ) 
    
  end

  def update_point_no_update( client, point )
    update_point( client, point, false )
  end
  
  def update_point( client, point, client_update = true )
    
    return unless point
    return unless player2screen( client.player ).know?( *point )

    showcell_radius( client, point, 0, client_update )

  end
  
  def hitmonster( client, monster, point )

    player = client.player
    mesg   = player.mesg

    return if player.timestop? # ֤ߤƤ֤ϹưԲġ

    # ̿Ψη׻
    hit = monster.ac + player.hitchance

    # Ϥ˱ƹ⤬ȡޤ2̤ꡣ
    
    if ( rnd( 20 ) < hit - $HARDGAME ) || ( rnd( 71 ) < 5 )
      message( client, mesg[ 440 ], player.monstername( monster ) )
      weapon_dulled( client, monster )
      monsterdamaged( client, monster, point, player.hitdamage( monster ) )
    else
      message( client, mesg[ 441 ], player.monstername( monster ) )
    end

  end

  # RogueȰ㤤ƥΰϤ򲼤󥹥򹶷⤹Ȥˤ
  # ΰϤ餷
  # ! Ȥäǡ󥹥αդӻäƤޤΤ?
  
  def weapon_dulled( client, monster )

    return unless monster.dulled?
    
    player = client.player
    mesg   = player.mesg

    player.weapon_dulled

    # ϤäȤˤΤΡ
    if ( player.wield? )
      message( client, mesg[ 442 ], player.monstername( monster ) )
      # beep
    end
    
  end

  def monster_killed( client, monster, point )

    @cavelevel.monster_killed( monster, point )
    update_point( client, point )

  end
  
  def vampire2bat( client, monster, point )

    # Хѥäơ礳֡
    @cavelevel.monster_killed( monster, point )
    @cavelevel.put_monster( Monster.bat, point )

    showcell_radius( client, point, 0 )
    
  end

  def monsterdamaged( client, monster, point, damage )
    
    player = client.player
    mesg   = player.mesg

    player.break_holdmonst
    
    monster.wakeup

    hp = monster.hp
    
    monster.damaged( damage )

    if monster.killed?
      
      monster_killed( client, monster, point ) # 饤ȤΡ
      player.monstkilled

      message( client, mesg[ 443 ], player.monstername( monster ) )
      raiseexperience( client, monster.experience ) # иͳ

      # ФޤƥФޤ
      dropgold( client, monster, point )
      dropsomething( client, monster, point )

      return hp
      
    elsif monster.change_to_bat?

      # ۷쵴ϡ礳ˤʤ餷
      vampire2bat( client, monster, point )

    end

    return damage
    
  end

  def put_object_near( client, object, point )

    d_point = @cavelevel.put_object_near( object, point )
    showcell_radius( client, d_point, 0 ) if d_point

  end

  
  def dropgold_amount( client, amount, point )
    put_object_near( client, Larnobject.dropgold( amount ), point )
  end

  
  def dropgold( client, monster, point )

    return unless monster.dropgold?
    dropgold_amount( client, monster.gold, point )
    
  end

  def dropsomething( client, monster, point )

    if monster.dropsomething?
      something( client, point, @level )
    elsif monster.leprechaun?
      creategem( client, point ) if ( rnd( 101 ) >= 75 )
      dropsomething( client, monster, point ) if ( rnd( 5 ) == 1 )
    end
    
  end

  def something( client, point, level )

    something( client, point, level ) if ( rnd( 101 ) < 8 )

    object = Larnobject.something( level )
    return if object.junk?

    d_point = @cavelevel.put_object_near( object, point )
    showcell_radius( client, d_point, 0 ) if d_point

  end

  def get_amount( client, mesg, cancel_mesg, debug_hash )
    client.get_amount( mesg, cancel_mesg, debug_hash )
  end

  def loadable?( client )
    @cavelevel.name2player( client.name )
  end
  
end
